# -*- coding: utf-8 -*-

import abc
import datetime as dt
from abc import ABCMeta, abstractmethod
import pandas as pd
import pathlib as pth
import typing as typ
import re
from docx2pdf import convert
import PyPDF2
import os

from pacifico_devel import Message
from pacifico_devel.api.src.util import Enumerations, Variant

from pacifico_devel.util.reportTemplate.src.core.Style.Style import Style
from pacifico_devel.util.reportTemplate.src.core.Plot.Plot import Plot
import pacifico_devel.util.reportTemplate.src.io.PlotFactory as pltf
import pacifico_devel.util.reportTemplate.src.util.Enumerations as renm
import pacifico_devel.util.reportTemplate.src.util.Report.DateFormatterEn as dtf

from pacifico_devel.util.lexikon.src.util.Enumerations import Language
from pacifico_devel.util.lexikon.src.core.Translator import Translator

from pacifico_devel.util.aws import S3
from pacifico_devel.util.cfg import Configuration as cfg

multiDate = typ.Union[dt.date, dt.datetime]


class Report(metaclass=ABCMeta):
    """
    Puts it all together? (format and style, analogue to HTML and CSS.)
    """

    _BASE_PLOT_RESOLUTION: int = 1000
    _EPSILON: float = 2 / 100

    def __init__(self,
                 pathTemplate: typ.Union[pth.Path, str],
                 languageOutput: typ.Union[Language, str] = Language.Spanish,
                 languageInput: typ.Union[Language, str] = Language.Spanish,
                 date: dt.datetime = dt.datetime.now(),
                 outputFormat: typ.Optional[str] = None,
                 plotBackend: renm.Backend = renm.Backend.ALTAIR):

        self._plotBackend = plotBackend
        self._pathTemplate: pth.Path = pth.Path(pathTemplate)
        self._checkPathTemplate()

        if outputFormat is None:
            self._outputFormat: str = pathTemplate.suffix
        else:
            self._outputFormat: str = outputFormat

        if isinstance(languageOutput, str):
            self._languageOutput = Language.fromDeeplString(languageOutput)
        else:
            self._languageOutput = languageOutput

        if isinstance(languageInput, str):
            self._languageInput = Language.fromDeeplString(languageInput)
        else:
            self._languageInput = languageInput

        if not self._pathTemplate.suffix == self._outputFormat:
            raise ValueError(f"File {self._pathTemplate} must be a {self._outputFormat} document.")

        self._date = date
        self._tempPlots: typ.Dict[str, Plot] = {}

    @property
    def languageOutput(self) -> Language:
        return self._languageOutput

    @property
    def languageInput(self) -> Language:
        return self._languageInput

    @abc.abstractmethod
    def render(self,
               pathOutput: typ.Union[str, pth.Path]) -> None:
        """
        Generates document and saves it locally.

        Args:
            pathOutput:

        Returns:

        """
        pass

    def export(self,
               applicationName: str) -> typ.List[Message.Report]:
        """
        Make a file from abstract reportTemplate and save it to a S3 stated in the filename.

        Args:
           applicationName:

        Returns (str): URL of document.

        """
        self.render(pathOutput=self._tempOutputPath)
        url = self._uploadAndGetUrl(filepathLocal=self._tempOutputPath)
        dateEffective = dt.datetime.now()
        message = self._getMessage(url=url,
                                   applicationName=applicationName,
                                   dateEffective=dateEffective)
        return [message]

    def exportPDF(self, applicationName: str, clean: bool=True):
        self.render(pathOutput=self._tempOutputPath)
        convert(self._tempOutputPath)
        pathPDF = pth.Path(self._tempOutputPath.parent) / (self._tempOutputPath.stem + ".pdf")
        if clean:
            self.cleanPDF(pathPDF)
        url = self._uploadAndGetUrl(filepathLocal=pathPDF)
        dateEffective = dt.datetime.now()
        message = self._getMessage(url=url,
                                   applicationName=applicationName,
                                   dateEffective=dateEffective)
        return [message]

    def cleanPDF(self, filePathReader: typ.Union[str, pth.Path]):
        fileName: str = self._pathTemplate.stem + "_temp1.pdf"
        filePathWriter = self._pathTemplate.parent / fileName
        writer = PyPDF2.PdfFileWriter()
        file = open(filePathReader, 'rb')
        reader = PyPDF2.PdfFileReader(file)
        for index, page in enumerate(reader.pages):
            texts = [text for text in page.extractText().split('\n') if not re.search("^\s*$", text)]
            if len(texts) > 2 or index == 0:
                writer.addPage(page)
        outputFile = open(filePathWriter, "wb")
        writer.write(outputFile)
        file.close()
        os.remove(filePathReader)
        outputFile.close()
        os.rename(filePathWriter, filePathReader)

    def _deleteAllTempPlots(self) -> None:
        """
        deletes all temporary plot images

        Returns:

        """
        [plot.deleteImageFile()
         for plot in self._tempPlots.values()]

    def _getMessage(self,
                    url: str,
                    applicationName: str,
                    dateEffective: dt.datetime) -> Message.Report:
        """
        Returns information in API format (Message.Report).

        Args:
            url:
            applicationName:
            dateEffective:

        Returns:

        """
        # document, chapter, section, subsection, paragraph = self.splitApplicationName(applicationName=applicationName)
        keys: typ.List[str] = ["document", "chapter", "section", "subsection", "paragraph"]
        values: typ.List[str] = self._splitApplicationName(applicationName=applicationName)

        if not 1 <= len(values) <= 4:
            msg1 = "Expecting name to be split into list of between 1 and 4 elements, "
            msg2 = f"but returned: {values}"
            raise ValueError(msg1 + msg2)

        values.append(self._outputFormat)  # put format last
        argDict: typ.Dict[str, typ.Union[str, dt.datetime, Variant.Variant]] = {k: v
                                                                                for k, v in zip(keys, values)}
        argDict["item"] = "report"
        argDict["dateEffective"] = dateEffective
        variant: Variant.Variant = Variant.Variant(value=url,
                                                   valueType=Enumerations.ValueType.ValueType_Url)
        print(f"URL: {url}.")
        argDict["variant"] = variant

        message: Message.Report = Message.Report(**argDict)
        return message

    @abstractmethod
    def _setString(self,
                   tag,
                   value: str) -> None:
        """
        Adds string to document.

        Args:
            tag:
            value:

        Returns:

        """
        pass

    @abstractmethod
    def setImage(self,
                 tag: str,
                 path: typ.Union[str, pth.Path],
                 width: typ.Optional[float] = None,
                 height: typ.Optional[float] = None) -> None:
        """
        Sets an image.

        Returns:

        """

        pass

    def setString(self,
                  tag: str,
                  value: str,
                  translate: bool = False) -> None:  # could be enumeration or beautifulSoup:
        """
        For bond mnemo (for example).

        Args:
            value:
            tag:
            translate:

        Returns:

        """
        valueString = value
        if translate:
            print("\n" + "#" * 20)
            print("TRANSLATING...")
            msg1 = f"TEXT TO TRANSLATE FROM {self._languageInput.deeplStringInput} to {self._languageOutput.deeplStringOutput}:"
            msg2 = f"\n\t{valueString}"
            print(msg1 + msg2)
            valueString = Translator.translate(text=valueString,
                                               languageInput=self.languageInput,
                                               languageOutput=self.languageOutput)
            print(f"TRANSLATION : \n\t{valueString}")
            print("TRANSLATING... DONE!")
            print("#" * 20 + "\n")
        self._setString(value=valueString,
                        tag=tag)

    def setText(self,
                tag: str,
                value: str):
        """
        For Commentary (for example)

        Args:
            value:
            tag:

        Returns:

        """
        return self.setString(value=value,
                              tag=tag,
                              translate=True)

    def setInteger(self,
                   tag: str,
                   value: int) -> None:
        self.setString(value=Style.convertIntToString(value,
                                                      self.languageOutput),
                       tag=tag)

    def setFloat(self,
                 tag: str,
                 value: float,
                 numberDecimals: int,
                 isPercentage: bool = False) -> None:
        """
        Same as above.

        Args:
            value:
            numberDecimals:
            tag:
            isPercentage:

        Returns:

        """
        valueString = Style.convertFloatToString(value=value,
                                                 numberDecimals=numberDecimals,
                                                 language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setLongDate(self,
                    tag: str,
                    value: typ.Optional[multiDate] = None,
                    ) -> None:

        if value is None:
            longDateStringEn = dtf.DateFormatterEn.getLongDateStringEn(self._date)
        else:
            longDateStringEn = dtf.DateFormatterEn.getLongDateStringEn(value)

        longDateString = Translator.translate(text=longDateStringEn,
                                              languageInput=Language.English_US,
                                              languageOutput=self.languageOutput)
        self.setString(value=longDateString,
                       tag=tag,
                       translate=False)

    def setDateTime(self,
                    value: dt.datetime,
                    tag: str) -> None:
        valueString = Style.convertDateTimeToString(value=value,
                                                    language=self.languageOutput)
        self.setString(value=valueString,
                       tag=tag)

    def setDate(self,
                value: dt.date,
                tag: str) -> None:
        """


        Args:
            value:
            tag:

        Returns:

        """
        valueString = Style.convertDateToString(value,
                                                language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setDateMonth(self,
                     value: dt.date,
                     tag: str) -> None:
        """
        YYYY/MM

        Args:
            value:
            tag:

        Returns:

        """
        valueString = Style.convertDateMonthToString(value,
                                                     language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setDateQuarter(self,
                       value: dt.date,
                       tag: str) -> None:
        """
        YYYY

        Args:
            value:
            tag:

        Returns:

        """
        valueString = Style.convertDateQuarterToString(value,
                                                       language=self.languageOutput)

        self.setString(value=valueString,
                       tag=tag)

    def setDateYear(self,
                    value: dt.date,
                    tag: str) -> None:
        """
        YYYY

        Args:
            value:
            tag:

        Returns:

        """
        valueString = Style.convertDateYearToString(value)

        self.setString(value=valueString,
                       tag=tag)

    def setPlot(self,
                tag: str,
                df: pd.DataFrame,
                plotKind: typ.Union[renm.PlotKind, str],
                title: str,
                shape: typ.Tuple[float, float],
                backend: typ.Union[renm.Backend, str, None] = None):  # width, height

        if backend is None:
            backend = self._plotBackend

        if isinstance(plotKind, str):
            plotKind = renm.PlotKind[plotKind.upper()]

        p = pltf.PlotFactory.getPlot(plotKind=plotKind,
                                     languageInput=self._languageInput,
                                     languageOutput=self._languageOutput)

        if title:
            titleTranslated = Translator.translate(title,
                                                   languageInput=self._languageInput,
                                                   languageOutput=self._languageOutput)
        else:
            titleTranslated = title

        p.makePlot(df=df,
                   title=titleTranslated,
                   shape=tuple(int(dim * self._BASE_PLOT_RESOLUTION)
                               for dim in shape
                               ),
                   backend=backend)
        self.setImage(tag=tag,
                      path=p.pathImage,
                      width=shape[0],
                      height=shape[1])
        self._tempPlots[tag] = p

    def setLinePlot(self,
                    tag: str,
                    df: pd.DataFrame,
                    title: str,
                    shape: typ.Tuple[float, float]) -> None:

        p = pltf.PlotFactory.getPlot(plotKind=renm.PlotKind.LINE_PLOT,
                                     languageInput=self._languageInput,
                                     languageOutput=self._languageOutput)

        if title:
            titleTranslated = Translator.translate(title,
                                                   languageInput=self._languageInput,
                                                   languageOutput=self._languageOutput)
        else:
            titleTranslated = title

        p.makePlot(df=df,
                   title=titleTranslated,
                   shape=tuple(int(dim * self._BASE_PLOT_RESOLUTION)
                               for dim in shape
                               ),
                   backend=self._plotBackend)
        self.setImage(tag=tag,
                      path=p.pathImage,
                      width=shape[0],
                      height=shape[1])
        self._tempPlots[tag] = p

    def setBarplot(self,
                   tag: str,
                   df: pd.DataFrame,
                   title: str,
                   shape: typ.Tuple[float, float]) -> None:
        p = pltf.PlotFactory.getPlot(plotKind=renm.PlotKind.BAR_PLOT,
                                     languageInput=self._languageInput,
                                     languageOutput=self._languageOutput)

        if title:
            titleTranslated = Translator.translate(title,
                                                   languageInput=self._languageInput,
                                                   languageOutput=self._languageOutput)
        else:
            titleTranslated = title

        p.makePlot(df=df,
                   title=titleTranslated,
                   shape=tuple(int(dim * self._BASE_PLOT_RESOLUTION)
                               for dim in shape
                               ),
                   backend=self._plotBackend)
        self.setImage(tag=tag,
                      path=p.pathImage,
                      width=shape[0],
                      height=shape[1])
        self._tempPlots[tag] = p

    def setTable(self,
                 df: pd.DataFrame,
                 tag: str,
                 shape: typ.Tuple[int, int] = (10, 10),
                 formatter: dict = None):
        """
        PLAN: df (Done through Style) -> HTML -> IMAGE -> Paste in reportTemplate.

        Args:
            df:
            tag:
            shape:

        Returns:

        """
        p = pltf.PlotFactory.getPlot(plotKind=renm.PlotKind.TABLE_PLOT,
                                     languageInput=self._languageInput,
                                     languageOutput=self._languageOutput)

        p.makePlot(df=df,
                   shape=tuple(int(dim * self._BASE_PLOT_RESOLUTION)
                               for dim in shape),
                   backend=self._plotBackend,
                   formatter=formatter)
        self.setImage(tag=tag,
                      path=p.pathImage,
                      width=shape[0],
                      height=shape[1])
        self._tempPlots[tag] = p

    @property
    def _tempOutputPath(self) -> pth.Path:
        fileName: str = self._pathTemplate.stem + "_temp" + self._outputFormat
        return self._pathTemplate.parent / fileName

    def _checkPathTemplate(self) -> None:
        """
        Assures the input template's path exists.

        """
        if not self._pathTemplate.is_file():
            msg = f"No such template path: {self._pathTemplate}. Please initialize with a valid template path."
            raise FileNotFoundError(msg)

    def _checkOuputFormat(self) -> None:
        """
        Assures input template's format is the output's file format.

        """
        if not self._pathTemplate.suffix == self._outputFormat:
            raise ValueError(f"File {self._pathTemplate} must be a {self._outputFormat} document.")

    def _uploadAndGetUrl(self, filepathLocal: typ.Union[str, pth.Path]) -> str:

        """
        Uploads file to S3 and returns URL.

        Args:
            filepathLocal:

        Returns:

        """

        if isinstance(filepathLocal, str):
            filepathLocal = pth.Path(filepathLocal)
        filepathBucket: pth.Path = self._getFilePathUpload(filePathLocal=filepathLocal)
        bucketName = "media-pacifico"
        awsAccessKey = cfg.get("AWS_ACCESS_KEY_PACIFICO")
        awsSecretKey = cfg.get("AWS_SECRET_KEY_PACIFICO")
        S3.writeFileToBucket(filepathLocal.as_posix(),
                             bucketName,
                             awsAccessKey,
                             awsSecretKey,
                             filepathBucket.as_posix())  # also erases file
        url = S3.getUrl(filepathBucket.as_posix(),
                        bucketName,
                        awsAccessKey=awsAccessKey,
                        awsSecretKey=awsSecretKey)
        return url

    @staticmethod
    def _splitApplicationName(applicationName: str) -> typ.List[str]:
        splitName: typ.List[str] = re.findall("[A-Z][a-z]+",
                                              applicationName)  # NEEDS TO INCLUDE CAPITAL LETTERS FOLLOWING
        splitName = [s for s in splitName if s != "Application"]
        if len(splitName) == 0:
            splitName = ["Report"]
        if len(splitName) >= 5:
            sFinal = " ".join(splitName[4:])
            splitName = splitName[:4]
            splitName[4] = sFinal
        return splitName

    @staticmethod
    def _getFilePathUpload(filePathLocal: pth.Path) -> pth.Path:
        stringTime: str = dt.datetime.now().strftime("%Y.%m.%d_%H.%M.%S.%f")
        fileNameUpload = filePathLocal.stem + "_" + stringTime + filePathLocal.suffix
        return pth.Path(fileNameUpload)


if __name__ == '__main__':
    # print(__file__)
    # print(pth.Path(pth.Path(__file__)))
    # test = "ApplicationStuffABTesting"
    # print(Report.splitApplicationName(test))
    filepathBucket = "template_temp_2022.04.22_17.17.36.906014.docx"
    filepathLocal = "template_temp_2022.04.22_17.17.36.906014.docx"
    bucketName = "media-pacifico"
    awsAccessKey = cfg.get("AWS_ACCESS_KEY_PACIFICO")
    awsSecretKey = cfg.get("AWS_SECRET_KEY_PACIFICO")
    # S3.writeFileToBucket(filepathLocal.as_posix(),
    #                      bucketName,
    #                      awsAccessKey,
    #                      awsSecretKey,
    #                      filepathBucket.as_posix())  # also erases file
    url = S3.getUrl(filepathBucket,
                    bucketName,
                    awsAccessKey=awsAccessKey,
                    awsSecretKey=awsSecretKey)
    print(url)



"""Copyright © 2022 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import requests
import json
from ...cfg import Configuration  # from pacifico_devel.util.cfg import Configuration


class OpenFIGI:
    idTypes = {
        'ISIN': 'ID_ISIN',
        'CUSIP': 'ID_CUSIP',
        'Ticker': 'TICKER',
        'FIGI': 'ID_BB_GLOBAL'
    }

    def __init__(self):
        self.endpoint = 'https://api.openfigi.com'
        self.apiKey = Configuration.get('OPENFIGI_API_KEY')
        self.headers = {
            'Content-Type': 'application/json',
            'X-OPENFIGI-APIKEY': self.apiKey
        }
        self.mappingEndpoint = self.endpoint + '/v3/mapping'
        self.searchEndpoint = self.endpoint + '/v3/search'

    def getEndpoint(self):
        return self.endpoint

    def getApiKey(self):
        return self.apiKey

    def getHeaders(self):
        return self.headers

    def getMappingEndpoint(self):
        return self.mappingEndpoint

    def getSearchEndpoint(self):
        return self.searchEndpoint

    @classmethod
    def getIDTypes(cls):
        return cls.idTypes

    def getIDType(self, key):
        return self.getIDTypes().get(key)

    def search(self, key):
        payload = self._buildSearchPayload(key)
        response = requests.post(self.getSearchEndpoint(), data=payload, headers=self.getHeaders()).text
        response = json.loads(response)
        return self.__parseResponse(response)

    def _buildSearchPayload(self, key):
        payload = {
            "query": key,
        }
        return json.dumps(payload)

    def map(self, ISINs=None, CUSIPs=None, Tickers=None, FIGIs=None):
        if ISINs is None:
            ISINs = []
        if CUSIPs is None:
            CUSIPs = []
        if Tickers is None:
            Tickers = []
        if FIGIs is None:
            FIGIs = []
        searchList = ISINs + CUSIPs + Tickers + FIGIs
        mappingDict = self._buildMappingDict(ISINs, CUSIPs, Tickers, FIGIs)
        payload = self._buildMappingPayload(mappingDict)
        response = requests.post(self.getMappingEndpoint(), data=payload, headers=self.getHeaders()).text
        response = json.loads(response)
        responseDict = {}
        for search, searchResponse in zip(searchList, response):
            response = self.__parseResponse(searchResponse)
            if search in Tickers and response is None:
                response = self.search(search)
            responseDict.update({search: response})
        return responseDict

    def _buildMappingDict(self, ISINs, CUSIPs, Tickers, FIGIs):
        if not isinstance(ISINs, list):
            ISINs = [ISINs]
        if not isinstance(CUSIPs, list):
            CUSIPs = [CUSIPs]
        if not isinstance(Tickers, list):
            Tickers = [Tickers]
        if not isinstance(FIGIs, list):
            FIGIs = [FIGIs]
        mappingDict = {
            'ISIN': ISINs,
            'CUSIP': CUSIPs,
            'Ticker': Tickers,
            'FIGI': FIGIs
        }
        return mappingDict

    def _buildMappingPayload(self, mappingDict):
        # '[{"idType":"ID_WERTPAPIER","idValue":"851399"}]'
        payloadList = []
        for idType, idList in mappingDict.items():
            for specificId in idList:
                idDict = {"idType": self.getIDType(idType), "idValue": specificId}
                payloadList.append(idDict)
        return json.dumps(payloadList)

    def __parseResponse(self, response):
        if isinstance(response, str):
            response = json.loads(response)
        if "data" not in response.keys():
            return None
        reponseList = response.get("data")
        if reponseList == []:
            return None
        validResponse = reponseList[0]  # Select first hit
        return OpenFIGIResponse.fromDict(validResponse)

class OpenFIGIResponse:
    def __init__(self, figi, name, ticker, exchangeCode, compositeFigi, securityType, marketSector, shareClassFigi,
                 securityType2, securityDescription):
        self.figi = figi
        self.name = name
        self.ticker = ticker
        self.exchangeCode = exchangeCode
        self.compositeFigi = compositeFigi
        self.securityType = securityType
        self.marketSector = marketSector
        self.shareClassFigi = shareClassFigi
        self.securityType2 = securityType2
        self.securityDescription = securityDescription

    def getFIGI(self):
        return self.figi

    def getName(self):
        return self.name

    def getTicker(self):
        return self.ticker

    def getExchangeCode(self):
        return self.exchangeCode

    def getCompositeFigi(self):
        return self.compositeFigi

    def getSecurityType(self):
        return self.securityType

    def getMarketSector(self):
        return self.marketSector

    def getShareClassFigi(self):
        return self.shareClassFigi

    def getSecurityType2(self):
        return self.securityType2

    def getSecurityDescription(self):
        return self.securityDescription

    def getFamily(self):
        instrumentType = "NA"
        if self.getMarketSector() == "Corp":
            instrumentType = "BE"
        elif self.getMarketSector() == "Govt":
            instrumentType = "BT"
        elif self.getMarketSector() == "Equity":
                instrumentType = "AC"
                if "Mutual Fund" in self.getSecurityType2():
                    instrumentType = "CFM"
                if "ETF" in self.getName() or "ETF" in self.getSecurityType() or "ETP" in self.getSecurityType():
                    instrumentType = "ETF"
        return instrumentType

    def getJurisdiction(self):
        countryCode = self.getExchangeCode()
        if countryCode == "US":
            return "USA"
        if countryCode in ["LU", "NL", "IE", "FR", "DE", "IT", "ES", "AT", "SE"]:
            return "EU"
        if countryCode == "GB":
            return "UK"
        if countryCode == "JP":
            return "Japan"
        if countryCode in ["BR", "BZ"]:
            return "Brazil"
        if countryCode == "SG":
            return "Singapore"
        if countryCode == "BS":
            return "Bahamas"
        if countryCode == "VG":
            return "British Virgin Islands"
        if countryCode == "BM":
            return "British Bermuda"
        if countryCode == "BS":
            return "Bahamas"
        if countryCode == "KY":
            return "Cayman Islands"
        if countryCode == "PA":
            return "Panama"
        if countryCode == "HK":
            return "Hong Kong"
        if countryCode == "KR":
            return "South Korea"
        if countryCode in ['SANTIAGO', 'CHI']:
            return 'Chile'
        if countryCode == "MULT":
            return "World"
        if countryCode is None:
            countryCode = "NA"
        return countryCode

    def getNumeraire(self):
        return ''

    @classmethod
    def fromDict(cls, dict):
        figi = dict.get("figi")
        name = dict.get("name")
        ticker = dict.get("ticker")
        exchangeCode = dict.get("exchCode")
        compositeFigi = dict.get("compositeFIGI")
        securityType = dict.get("securityType")
        marketSector = dict.get("marketSector")
        shareClassFigi = dict.get("shareClassFIGI")
        securityType2 = dict.get("securityType2")
        securityDescription = dict.get("securityDescription")
        return cls(figi, name, ticker, exchangeCode, compositeFigi, securityType, marketSector, shareClassFigi, securityType2, securityDescription)

if __name__ == '__main__':
    openFIGI = OpenFIGI()
    ISINs = ['US29081P3038', 'US4592001014', 'CL0002614940', 'caca']
    tickers = ['AAPL', 'TSLA', 'SANTANDER', 'SPY']
    y = openFIGI.map(ISINs, None, tickers)
    print(y)
    for x in y.values():
        if x is not None:
            print(x.__dict__)



"""Copyright © 2022 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import json
import pandas
import datetime
import requests
from .. import Strings, Dates  # from pacifico_devel.util import Strings, Dates
from ..aws import FTP  # from pacifico_devel.util.aws import FTP
from ..cfg import Configuration  # from pacifico_devel.util.cfg import Configuration
from .OpenFIGI import OpenFIGI  # from pacifico_devel.util.spellChecker.OpenFIGI import OpenFIGI

class SpellChecker:
    def __init__(self):
        self.openFIGI = OpenFIGI.OpenFIGI()
        self.ftpPath = 'backup/input/datawarehouse/sanexpedito/'

    def getOpenFIGI(self):
        return self.openFIGI

    def getFTPPath(self):
        return self.ftpPath

    def isISIN(self, key: str):
        key = key.upper()
        if len(key) != 12:
            return False
        elif not key[0].isalpha() or not key[1].isalpha():
            return False
        elif not all([character.isalnum() for character in key[2:]]):
            return False
        else:
            return True

    def isCUSIP(self, key: str):
        key = key.upper()
        if len(key) != 9:
            return False
        elif not all([character.isnumeric() for character in key[:3]]):
            return False
        elif not all([character.isalnum() for character in key[3:8]]):
            return False
        elif not key[-1].isnumeric():
            return False
        else:
            return True

    def isFIGI(self, key: str):
        key = key.upper()
        if len(key) != 12:
            return False
        elif not all([character.isalpha() for character in key[:2]]):
            return False
        elif key[2] != 'G':
            return False
        elif not all([character.isalnum() for character in key[3:11]]):
            return False
        elif not key[-1].isnumeric():
            return False
        else:
            return True

    def classifyTickers(self, tickers, verbose=False):
        ISINs = []
        CUSIPs = []
        Tickers = []
        FIGIs = []
        for ticker in tickers:
            isISIN = self.isISIN(ticker)
            isCUSIP = self.isCUSIP(ticker)
            isFIGI = self.isFIGI(ticker)
            if verbose:
                print(f'{ticker}: {isISIN = } / {isCUSIP = } / {isFIGI = }')
            if isISIN:
                ISINs.append(ticker)
            elif isCUSIP:
                CUSIPs.append(ticker)
            elif isFIGI:
                FIGIs.append(ticker)
            else:
                Tickers.append(ticker)
        return ISINs, CUSIPs, Tickers, FIGIs

    def cleanTickers(self, tickers):
        # Standardize the list
        if isinstance(tickers, str):
            if '[' in tickers:
                tickers = json.loads(tickers)
            else:
                tickers = [tickers]
        # Actually clean the values
        cleanTickers = []
        for ticker in tickers:
            ticker = Strings.cleanString(ticker).strip()
            cleanTickers.append(ticker)
        return cleanTickers

    def checkForAlreadySearched(self, spellingDict, tickers):
        newTickers = []
        existing = {}
        for ticker in tickers:
            alreadySearched = spellingDict.get(ticker)
            if alreadySearched is None:
                newTickers.append(ticker)
            else:
                existing.update({ticker: alreadySearched})
        return newTickers, existing

    def addToRecovered(self, responses):
        rows = []
        addedFigis = []
        recoveredFile = self.getRecoveredFile()
        dateNow = Dates.getExcelDate(datetime.datetime.now())
        for key, response in responses.items():
            if response is not None:
                family = response.getFamily()
                jurisdiction = response.getJurisdiction()
                ticker = response.getTicker()
                numeraire = response.getNumeraire()
                figi = response.getFIGI()
                if figi not in addedFigis and ticker not in recoveredFile['Ticker BBG'].values:
                    isISIN = self.isISIN(key)
                    if isISIN:
                        ISIN = key
                    else:
                        ISIN = 'NA'
                    rows.append({'Mnemo': ISIN, 'Family': family, 'Jurisdiction': jurisdiction, 'Ticker BBG': ticker,
                                 'Date': dateNow, 'Numeraire BBG': numeraire, 'FIGI': figi})
                    addedFigis.append(figi)
            else:
                rows.append({'Mnemo': key, 'Family': '', 'Jurisdiction': '', 'Ticker BBG': 'not found', 'Date': dateNow,
                             'Numeraire BBG': '', 'FIGI': ''})
        dataframe = pandas.DataFrame(rows)
        newRecoveredFile = pandas.concat([recoveredFile, dataframe], ignore_index=True)
        # print(newRecoveredFile)
        self.updateRecoveredFile(newRecoveredFile)

    def getRecoveredFile(self):
        # unavailable = FTP.getCSVFileAsPandas(self.getFTPPath() + 'unavailable.csv', tmp=False)
        recovered = FTP.getCSVFileAsPandas(self.getFTPPath() + 'recovered.csv', tmp=False)
        return recovered

    def updateRecoveredFile(self, recoveredDataframe):
        FTP.uploadPandasAsCSV(recoveredDataframe, self.getFTPPath() + 'recovered.csv')

    def getSpellingFile(self):
        spelling = FTP.getFileAsString(self.getFTPPath() + 'spelling.json')
        return json.loads(spelling)

    def updateSpellingFile(self, spellingDict):
        spellingJson = json.dumps(spellingDict)
        FTP.uploadFileFromString(spellingJson, self.getFTPPath() + 'spelling.json')

    def runSanExpeditoAdmin(self):
        appRunnerBotsEndpoint = Configuration.get('APP_RUNNER_ENDPOINT')
        headers = {'Content-Type': 'application/json'}
        payload = {
            "bot name": "BotSanExpedito",
            "arguments": {
                "executionType": "admin"
            }
        }
        payload = json.dumps(payload)
        response = requests.post(appRunnerBotsEndpoint, data=payload, headers=headers).text
        print(f'SpellChecker: runSanExpeditoAdmin response -> {response}')

    def searchNewTickers(self, tickers, runSEAdmin):
        addedDict = {}
        # Classify tickers
        ISINs, CUSIPs, Tickers, FIGIs = self.classifyTickers(tickers)
        # Get OpenFIGI responses
        responses = self.getOpenFIGI().map(ISINs, CUSIPs, Tickers, FIGIs)
        for key, response in responses.items():
            if response is not None:
                # if key in Tickers and key != response.getTicker():
                #     print(f'Did you mean {response.getTicker()} ({response.getName()})')
                addedDict.update({key: (response.getTicker(), response.getName())})
        self.addToRecovered(responses)
        if runSEAdmin:
            self.runSanExpeditoAdmin()
        return addedDict

    def search(self, tickers, runSEAdmin=True):
        # Clean tickers
        tickers = self.cleanTickers(tickers)
        # Get spelling file
        spellingDict = self.getSpellingFile()
        # Check for already existing ones
        tickers, existingDict = self.checkForAlreadySearched(spellingDict, tickers)
        if tickers != []:
            addedDict = self.searchNewTickers(tickers, runSEAdmin)
            for ticker, added in addedDict.items():
                tickerFound, nameFound = added
                spellingDict.update({ticker: tickerFound})
        else:
            addedDict = {}
        # print(spellingDict)
        self.updateSpellingFile(spellingDict)
        return existingDict, addedDict

if __name__ == '__main__':
    pandas.set_option('display.max_rows', None)
    pandas.set_option('display.max_columns', None)
    pandas.set_option('display.width', 1000)
    pandas.set_option('display.colheader_justify', 'center')
    pandas.set_option('display.precision', 3)
    spellChecker = SpellChecker()
    tickers = 'US29081P3038'  # ['US4592001014', 'CL0002614940']#, 'caca', 'AAPL', 'TSLA', 'SANTANDER', '037833100', '38259P508',
               # 'BBG000B9Y2J5', 'BBG000007X44']
    existingDict, addedDict = spellChecker.search(tickers)
    print(existingDict)
    print(addedDict)


"""Copyright © 2022 Burrus Financial Intelligence, Ltda. (hereafter, BFI) Permission to include in application
software or to make digital or hard copies of part or all of this work is subject to the following licensing
agreement.
BFI Software License Agreement: Any User wishing to make a commercial use of the Software must contact BFI
at jacques.burrus@bfi.lat to arrange an appropriate license. Commercial use includes (1) integrating or incorporating
all or part of the source code into a product for sale or license by, or on behalf of, User to third parties,
or (2) distribution of the binary or source code to third parties for use with a commercial product sold or licensed
by, or on behalf of, User. """

import time
import multiprocessing
import datetime
import pandas
from enum import Enum
from pacifico_devel.util.aws import EC2, S3  # from ..aws import EC2, S3
from pacifico_devel.util.cfg import Configuration  # from ..cfg import Configuration
from pacifico_devel.util import Dates  # from .. import Dates

class Status(Enum):
    Started = 0
    Ended = 1
    Failed = 2

    def getString(self):
        return self.name

    @classmethod
    def fromString(cls, string):
        return cls[string]

    def isFinished(self):
        if self in [Status.Ended, Status.Failed]:
            return True
        return False

class LoadBalancer:
    cooldownInSeconds = 180
    errorInSeconds = 10

    @staticmethod
    def getTimestamp():
        dateNow = datetime.datetime.now()
        timestamp = Dates.getDateTimeTimestamp(dateNow)
        return timestamp

    @staticmethod
    def getTimestampId():
        dateNow = datetime.datetime.now()
        timestampExcel = Dates.getExcelDate(dateNow)
        timestampId = int(timestampExcel * 1000000000)
        return timestampId

    @staticmethod
    def _checkIfBiggerInList(value, list):
        for x in list:
            if value >= x:
                return False
        return True

    @staticmethod
    def stopInstance(instanceId, awsAccessKey, awsSecretKey, force=False):
        computerStaticInstanceId = Configuration.get('COMPUTER_STATIC_2_INSTANCE_ID')
        if instanceId != computerStaticInstanceId or force:
            EC2.stopInstace(instanceId, awsAccessKey, awsSecretKey)


    @staticmethod
    def notifyStart(instanceId):
        processId = LoadBalancer.getTimestampId()
        t = multiprocessing.Process(target=LoadBalancer._notifyStart, args=(processId, instanceId,), daemon=False)
        t.start()
        return processId

    @staticmethod
    def _notifyStart(processId, instanceId):
        scheduleFile = LoadBalancer.getScheduleFile()
        timestamp = LoadBalancer.getTimestamp()
        status = Status.Started.getString()
        # Add start to schedule (invalidate previous stops)
        scheduleFile = LoadBalancer._updateOrAddRowToFile(scheduleFile, processId, instanceId, timestamp, status)
        # Clean older than a month
        scheduleFile = LoadBalancer.cleanScheduleFile(scheduleFile)
        # Update schedule file
        LoadBalancer.uploadScheduleFile(scheduleFile)
        print(f'notifyStart : {processId, instanceId}')
        print(scheduleFile)
        print(f'notifyStart {"*" * 10}')

    @staticmethod
    def notifyStop(processId, instanceId):
        t = multiprocessing.Process(target=LoadBalancer._notifyStop, args=(processId, instanceId,), daemon=False)
        t.start()

    @staticmethod
    def notifyEnd(processId, instanceId):
        t = multiprocessing.Process(target=LoadBalancer._notifyEnd, args=(processId, instanceId,), daemon=False)
        t.start()

    @staticmethod
    def notifyFailure(processId, instanceId):
        t = multiprocessing.Process(target=LoadBalancer._notifyFailure, args=(processId, instanceId,), daemon=False)
        t.start()

    @staticmethod
    def _notifyStop(processId, instanceId):
        LoadBalancer._notifyEnd(processId, instanceId)
        # Wait the cooldown period
        time.sleep(LoadBalancer.cooldownInSeconds)
        # Actually shutdown if it's valid
        LoadBalancer.executeStop(instanceId)
        print(f'notifyStop : {processId, instanceId}')

    @staticmethod
    def _notifyEnd(processId, instanceId):
        scheduleFile = LoadBalancer.getScheduleFile()
        timestamp = LoadBalancer.getTimestamp()
        status = Status.Ended.getString()
        # Add stop to schedule
        scheduleFile = LoadBalancer._updateOrAddRowToFile(scheduleFile, processId, instanceId, timestamp, status)
        # Update schedule file
        LoadBalancer.uploadScheduleFile(scheduleFile)
        print(f'notifyEnd : {processId, instanceId}')
        print(scheduleFile)
        print(f'notifyEnd {"*" * 10}')

    @staticmethod
    def _notifyFailure(processId, instanceId):
        scheduleFile = LoadBalancer.getScheduleFile()
        timestamp = LoadBalancer.getTimestamp()
        status = Status.Failed.getString()
        # Add stop to schedule
        scheduleFile = LoadBalancer._updateOrAddRowToFile(scheduleFile, processId, instanceId, timestamp, status)
        # Update schedule file
        LoadBalancer.uploadScheduleFile(scheduleFile)
        print(f'notifyFailure : {processId, instanceId}')
        print(scheduleFile)
        print(f'notifyFailure {"*" * 10}')

    @staticmethod
    def executeStop(instanceId):
        awsAccessKey = Configuration.get('AWS_ACCESS_KEY')
        awsSecretKey = Configuration.get('AWS_SECRET_KEY')
        # Check schedule
        instanceIsOn = EC2.isInstanceRunning(instanceId, awsAccessKey, awsSecretKey)
        if instanceIsOn:
            try:
                scheduleFile = LoadBalancer.getScheduleFile()
                relevantRows = scheduleFile.loc[scheduleFile['instanceId'] == instanceId]
                status = relevantRows['status'].unique()
                noOtherProcessesRunning = Status.Started.getString() not in status
                timestamps = relevantRows['timestamp'].unique()
                timedeltas = [datetime.datetime.now() - Dates.getDateTimeFromTimestamp(timestamp) for timestamp in timestamps]
                coolDownDelta = datetime.timedelta(seconds=(LoadBalancer.cooldownInSeconds - LoadBalancer.errorInSeconds))
                coolDownEnded = LoadBalancer._checkIfBiggerInList(coolDownDelta, timedeltas)
                isShutdownValid = noOtherProcessesRunning and coolDownEnded
                print(f'isShutdownValid: {isShutdownValid}')
                if isShutdownValid:  # If valid, shutdown
                    LoadBalancer.stopInstance(instanceId, awsAccessKey, awsSecretKey)
            except Exception as e:
                print(f'Error with schedule file: {str(e)}')
                LoadBalancer.stopInstance(instanceId, awsAccessKey, awsSecretKey)  # If file fails just stop the instance

    @staticmethod
    def _updateOrAddRowToFile(scheduleFile, processId, instanceId, timestamp, status):
        if instanceId in scheduleFile['instanceId'].values and processId in scheduleFile['processId'].values:
            scheduleFile.loc[scheduleFile['instanceId'] == instanceId, 'timestamp'] = timestamp
            scheduleFile.loc[scheduleFile['instanceId'] == instanceId, 'status'] = status
        else:
            row = LoadBalancer.buildScheduleRow(processId, instanceId, timestamp, status)
            scheduleFileNewRows = pandas.DataFrame([row])
            scheduleFile = pandas.concat([scheduleFile, scheduleFileNewRows], ignore_index=True)
        return scheduleFile

    @staticmethod
    def buildScheduleRow(processId, instanceId, timestamp, status):
        return {'processId': processId, 'instanceId': instanceId, 'timestamp': timestamp, 'status': status}

    @staticmethod
    def cleanScheduleFile(scheduleFile):
        aMonthAgo = datetime.date.today() - datetime.timedelta(days=30)
        scheduleFile['timestamp'] = scheduleFile['timestamp'].apply(lambda timestamp: Dates.getDateTimeFromTimestamp(timestamp))
        scheduleFile = scheduleFile.loc[scheduleFile['timestamp'] >= pandas.to_datetime(aMonthAgo)]
        scheduleFile['timestamp'] = scheduleFile['timestamp'].apply(lambda timestamp: Dates.getDateTimeTimestamp(timestamp))
        return scheduleFile

    # Downloading/Uploading Files

    @staticmethod
    def __getScheduleFilePaths():
        localPath = 'schedule.csv'
        bucketPath = f'LoadBalancer/{localPath}'
        bucket = 'bfi-cfg'
        return localPath, bucketPath, bucket

    @staticmethod
    def getScheduleFile():
        localPath, bucketPath, bucket = LoadBalancer.__getScheduleFilePaths()
        awsAccessKey = Configuration.get('AWS_ACCESS_KEY')
        awsSecretKey = Configuration.get('AWS_SECRET_KEY')
        S3.downloadFromBucket(bucketPath, localPath, bucket, read=False, awsAccessKey=awsAccessKey, awsSecretKey=awsSecretKey)
        schedule = pandas.read_csv(localPath)
        return schedule

    @staticmethod
    def uploadScheduleFile(schedule):
        localPath, bucketPath, bucket = LoadBalancer.__getScheduleFilePaths()
        awsAccessKey = Configuration.get('AWS_ACCESS_KEY')
        awsSecretKey = Configuration.get('AWS_SECRET_KEY')
        schedule = schedule.to_csv(index=False, sep=',', line_terminator='\n')
        S3.writeToBucket(schedule, localPath, bucket, bucketFilename=bucketPath, awsAccessKey=awsAccessKey, awsSecretKey=awsSecretKey)
